#include "rf_char.ih"
#include "rf_expr.ih"
#include "rf_types.ih"
#include "rf_short_int.ih"
#include "rf_integer.ih"

namespace rftype
{

Expr Char::create_expr (const WString& _wstr)
{
  size_t len = _wstr.get_length();
  const wchar_t* pstr = _wstr.get_data();

  Expr  e = Term::create_flat_expr(len);
  Term* p = e.get_first();

  while (len--)
    new(p++) Char(*pstr++);
  return e;
}

Expr Char::create_expr (wchar_t const* _wstr, size_t _len /* = (size_t)-1 */)
{
  size_t len = _len == (size_t)-1 ? wcslen(_wstr) : _len;

  Expr  e = Term::create_flat_expr(len);
  Term* p = e.get_first();

  while (len--)
    new(p++) Char(*_wstr++);
  return e;
}

Expr Char::create_expr (const wchar_t _ch)
{
  Expr  e = Term::create_flat_expr(1);
  Term* p = e.get_first();
  new(p) Char(_ch);
  return e;
}

Expr Char::create_expr (char const* _str, char const* _locale /* = null */)
{
  char const* saved_locale = null;
  if (_locale != null) {
    saved_locale = setlocale(LC_CTYPE, null);
    setlocale(LC_CTYPE, _locale);
  }
  char const* s = _str;
  size_t len = mbstowcs(null, s, 0);
  if (len == (size_t)(-1)) FATAL("Error while converting a string");
  Expr e = Term::create_flat_expr(len);
  Term* p = e.get_first();
  wchar_t wc;
  while (len-- > 0) {
    size_t l = mbtowc(&wc, s, MB_CUR_MAX);
    new(p) Char(wc);
    p++; s += l;
  }
  if (_locale != null) setlocale(LC_CTYPE, saved_locale);
  return e;
}

Expr Char::to_chars (const Expr& _expr)
{
  static const Expr lp = Char::create_expr("(");
  static const Expr rp = Char::create_expr(")");

  Expr e;

  Term* p = _expr.get_first();
  Term* q = _expr.get_last();

  for (; p < q; p++)
  {
    if (p->is_ref())
      e = e + lp + to_chars(*p->get_obj_ptr<Expr>()) + rp;
    else
      e = e + Char::create_expr(WString(*p));
  };

  return e;
}

Expr Char::to_lower (const Expr& _expr)
{
  Term* p = _expr.get_first();
  Term* q = _expr.get_last();

  Expr  e = Term::create_expr(_expr.get_len());
  Term* r = e.get_first();
  
  for (; p < q; p++, r++)
    new(r) Char(towlower(((Char*)p)->uint_data2));

  return e;
}

Expr Char::to_upper (const Expr& _expr)
{
  Term* p = _expr.get_first();
  Term* q = _expr.get_last();

  Expr  e = Term::create_expr(_expr.get_len());
  Term* r = e.get_first();
  
  for (; p < q; p++, r++)
    new(r) Char(towupper(((Char*)p)->uint_data2));

  return e;
}

Expr Char::bytes_to_chars (const Expr& _expr)
{
  Term* p = _expr.get_first();
  Term* q = _expr.get_last();

  Expr  e = Term::create_expr(_expr.get_len());
  Term* r = e.get_first();
  
  uintptr_t temp;
  for (; p < q; p++, r++) {
    if (p->get_type() == type_int32)
      temp = static_cast<Int32*>(p)->to_int();
    else
      temp = static_cast<Integer*>(p)->to_int();
    new(r) Char(temp);
  };
  return e;
}

Expr Char::chars_to_bytes (const Expr& _expr)
{
  Term* p = _expr.get_first();
  Term* q = _expr.get_last();

  Expr  e = Term::create_expr(_expr.get_len());
  Term* r = e.get_first();
  for (; p < q; p++, r++) {
//    new(r) Int32(((Char*)p)->uint_data2);
    new(r) Integer(((Char*)p)->uint_data2);
  }
  return e;
}

pxx::WString Char::to_string (Char const* _char)
{
  wchar_t wc = _char->uint_data2;
  return pxx::WString(&wc, 1);
}

short_type_funcs_t Char::funcs = {
  0,
  null, // Char::ctor,
  null, // Char::dtor,
  null, // Char::eq,
  null, // (int (*)(Term const*, Term const*))(Char::compare),
  null, // (uint32_t (*)(Term const*))(Char::hash),
  (pxx::WString (*)(Term const*))(Char::to_string),
  null
};

}
